#include "../stdafx.h"
#include "TsDataMgr.h"
#include "../Common/FileRecord.h"
#include <shlwapi.h>

// ZLIB
#include "../lib/zlib.h"
#pragma comment(lib,"zlib.lib")

#pragma comment(lib,"shlwapi.lib")

CTsDataMgr::CTsDataMgr(void)
{
	CLock::CreateLockObj(&m_DataLock);
	m_pPayloadTemp = new BYTE[TS_MAX_PAYLOAD_LENGTH];
	m_lpFunc = NULL;
	m_lpContext = NULL;
#ifdef TSDATAMGR_DEBUG_TRACE_MEMORY
	m_dwDownloadBufferSize = 0;
	m_dwContentBufferSize = 0;
#endif
}

CTsDataMgr::~CTsDataMgr(void)
{
	Clear();	
	delete []m_pPayloadTemp;
	CLock::DestroyLockObj(&m_DataLock);
}

// Sf[^̃NA
void CTsDataMgr::Clear()
{
	CLock cDataLock(&m_DataLock);

	TsDownloadData::iterator it;
#ifdef TSDATAMGR_DEBUG_PRINT
		{
			char buf[80];
			wsprintfA(buf,"CTsDataMgr::Clear()\n");
			OutputDebugStringA(buf);
#ifdef TSDATAMGR_DEBUG_TRACE_MEMORY
			wsprintfA(buf,"\tDownload Buffer=%9d byte\n",m_dwDownloadBufferSize);
			OutputDebugStringA(buf);
			wsprintfA(buf,"\tContent  Buffer=%9d byte\n",m_dwContentBufferSize);
			OutputDebugStringA(buf);
#endif
		}
#endif
	for(it=m_DownloadData.begin();it!=m_DownloadData.end();it++)
	{
		ClearByPID(it->first,false);
	}
	m_DownloadData.clear();
	m_Content.clear();
}

// wPIDɏf[^폜
void CTsDataMgr::ClearByPID(WORD wPID,bool bRemoveItem)
{
	CLock cDataLock(&m_DataLock);
	if(m_DownloadData.size()>0)
	{
		TsDownloadData::iterator it = m_DownloadData.find(wPID);
		if(it!=m_DownloadData.end())
		{
			BYTE byComponentTag = it->second.byComponentTag;
#ifdef TSDATAMGR_DEBUG_PRINT
			{
				char dbg_buf[80];
				wsprintfA(dbg_buf,"%sCTsDataMgr::ClearByPID(PID=0x%04X,ComponentTag=%02X)\n",bRemoveItem?"":"\t",wPID,it->second.byComponentTag);
				OutputDebugStringA(dbg_buf);
			}
#endif

			// _E[hJ
			TAG_DownloadData &stDLData = it->second;
			for(int i=0;i<stDLData.wNumberOfModules;i++)
			{
				// _E[hpJ
				DownloadBuffer_Free(stDLData.Module[i].Status.pDownload);
			}
			// RecJ
			std::vector<TsContentLocation::iterator> vDelete;
			TsContentLocation::iterator itLocation;
			for(itLocation=m_Content.begin();itLocation!=m_Content.end();itLocation++)
			{
				if(CompareComponentTag(itLocation->first,it->second.byComponentTag))
				{
					// JR|[lg
#ifdef TSDATAMGR_DEBUG_PRINT
						{
							char dbg_buf[80];
							wsprintfA(dbg_buf,"%sDEL ModNo=0x%04X\n",bRemoveItem?"":"\t",itLocation->first&0xFFFF);
							OutputDebugStringA(dbg_buf);
						}
#endif
					TsContent::iterator itContent;
					for(itContent=itLocation->second.begin();itContent!=itLocation->second.end();itContent++)
					{
						ContentBuffer_Free(itContent->second.pContent);
#ifdef TSDATAMGR_DEBUG_PRINT_DETAIL
						{
							char dbg_buf[80];
							wsprintf(dbg_buf,"%s\tDelete Content \'%s\'\n",bRemoveItem?"":"\t",itContent->second.szPath);
							OutputDebugString(dbg_buf);
						}
#endif
					}
					itLocation->second.clear();
					vDelete.push_back(itLocation);
				}
			}
			if(bRemoveItem)
			{
				// 菜
				// RecXg폜
				for(unsigned int iDel=0;iDel<vDelete.size();iDel++)
				{
					m_Content.erase(vDelete[iDel]);
				}
				m_DownloadData.erase(it);
			} else {
				// [NA
				ZeroMemory(&stDLData,sizeof(stDLData));
			}
			// R|[lg폜ʒm
			DoCallback(TSENG_CALLBACK_DATA_COMPONENT_CLEAR,wPID,(LPVOID)byComponentTag);
		}
	}
}

// DSM-CC(DownloadInfoIndication TableID=0x3B)̃pPbg
bool CTsDataMgr::OnDII(WORD wPID,BYTE *pSection,int iPos,int iRemain,BYTE byComponentTag)
{
	CLock cDataLock(&m_DataLock);

	// MPEG-2 DSM-CC
	if(pSection[iPos+0]==0x11 && pSection[iPos+1]==0x03 && ( pSection[iPos+2]==0x10 && pSection[iPos+3]==0x02 ))
	{
		TAG_DownloadData stDLData;
		ZeroMemory(&stDLData,sizeof(stDLData));
		stDLData.byComponentTag = byComponentTag;

		stDLData.dwTransactionID = (DWORD)(pSection[iPos+4])<<24 | (DWORD)(pSection[iPos+5])<<16 | (DWORD)(pSection[iPos+6])<<8 | pSection[iPos+7];
		stDLData.dwTransactionNo = stDLData.dwTransactionID & 0x3FFFFFFF;
		BYTE byAdaptationLength = pSection[iPos+9];
		WORD wMessageLength = (WORD)(pSection[iPos+10])<<8 | pSection[iPos+11];
		
		iPos+=(byAdaptationLength+12);
		iRemain-=(byAdaptationLength+12);

		stDLData.dwDownloadID = (DWORD)(pSection[iPos+0])<<24 | (DWORD)(pSection[iPos+1])<<16 | (DWORD)(pSection[iPos+2])<<8 | pSection[iPos+3];
		stDLData.wBlockSize = (WORD)(pSection[iPos+4])<<8 | pSection[iPos+5];
		stDLData.byWindowSize = pSection[iPos+6];
		stDLData.byAckPriod = pSection[iPos+7];
		stDLData.dwTcDownloadWindow = (DWORD)(pSection[iPos+8])<<24 | (DWORD)(pSection[iPos+9])<<16 | (DWORD)(pSection[iPos+10])<<8 | pSection[iPos+11];
		stDLData.dwTcDownloadScenario = (DWORD)(pSection[iPos+12])<<24 | (DWORD)(pSection[iPos+13])<<16 | (DWORD)(pSection[iPos+14])<<8 | pSection[iPos+15];
		BYTE byDescType = pSection[iPos+16];
		BYTE byDescLen = pSection[iPos+17];
		stDLData.wNumberOfModules = (WORD)(pSection[iPos+byDescLen+18])<<8 | pSection[iPos+byDescLen+19];
		if(stDLData.wNumberOfModules>TSDOWNLOAD_MODULE_MAX)
		{
			// W[ző吔𒴂Ă邪Aȗ
			stDLData.wNumberOfModules=TSDOWNLOAD_MODULE_MAX;
		}
		
		iPos+=(byDescLen+20);
		iRemain-=(byDescLen+20);

		for(int iModule=0;iModule<stDLData.wNumberOfModules;iModule++)
		{
			stDLData.Module[iModule].wID = (WORD)(pSection[iPos+0])<<8 | pSection[iPos+1];
			stDLData.Module[iModule].dwSize = (DWORD)(pSection[iPos+2])<<24 | (DWORD)(pSection[iPos+3])<<16 | (DWORD)(pSection[iPos+4])<<8 | pSection[iPos+5];
			stDLData.Module[iModule].byVersion = pSection[iPos+6];
			stDLData.Module[iModule].wBlockNum = (WORD)( ((stDLData.Module[iModule].dwSize-1)/stDLData.wBlockSize) + 1);
			if(stDLData.Module[iModule].wBlockNum>TSDOWNLOAD_BLOCK_MAX)
			{
				// W[ző吔𒴂Ă邪Aȗ
				stDLData.Module[iModule].wBlockNum = TSDOWNLOAD_BLOCK_MAX;
			}
			BYTE byModuleInfoLen = pSection[iPos+7];
			for(BYTE byInfoPos=0;byInfoPos<byModuleInfoLen;)
			{
				BYTE byTag = pSection[iPos+8+byInfoPos+0];
				BYTE byLen = pSection[iPos+8+byInfoPos+1];
				switch(byTag)
				{
				case 0x01:
					{
						// t@C^Cvw
					}
					break;
				case 0xC2:
					{
						// kw
						stDLData.Module[iModule].Compress.bFlg = true;
						stDLData.Module[iModule].Compress.byType = pSection[iPos+8+byInfoPos+2];
						stDLData.Module[iModule].Compress.dwSize = (DWORD)(pSection[iPos+8+byInfoPos+3])<<24 | (DWORD)(pSection[iPos+8+byInfoPos+4])<<16 | (DWORD)(pSection[iPos+8+byInfoPos+5])<<8 | pSection[iPos+8+byInfoPos+6];
					}
					break;
				case 0xC3:
					{
						// Rg[ݒ
					}
					break;
				case 0xCA:
					{
						// [gؖݒ
					}
					break;
				}
				byInfoPos+=(byLen+2);
			}
			iPos+=(byModuleInfoLen+8);
			iRemain-=(byModuleInfoLen+8);
		}
		WORD wPrivateDataLength = (WORD)(pSection[iPos+0])<<8 | pSection[iPos+1];		
		
		bool bUpdateFlg = false;
		TsDownloadData::iterator it = m_DownloadData.find(wPID);
		if(it==m_DownloadData.end())
		{
			// ̂܂ܐVKǉ
			m_DownloadData.insert(TsDownloadDataPair(wPID,stDLData));
			bUpdateFlg = true;
			// XVʒm
			for(int iModule=0;iModule<stDLData.wNumberOfModules;iModule++)
			{
				TAG_TSCB_MODULE stNotify;
				stNotify.byComponentTag = byComponentTag;
				stNotify.wModuleID = stDLData.Module[iModule].wID;
				stNotify.byVersion = stDLData.Module[iModule].byVersion;
				stNotify.dwSize = stDLData.Module[iModule].dwSize;
				DoCallback(TSENG_CALLBACK_DATA_MODULE_UPDATE,wPID,&stNotify);
			}
		} else {
			// DIIƐVDII̓ID̃o[WrAXV
			TAG_DownloadData &stDLOld = it->second;
			for(int iOld=0;iOld<stDLOld.wNumberOfModules;iOld++)
			{
				bool bFindFlg = false;
				bool bFreeFlg = true;
				for(int iNew=0;!bFindFlg && iNew<stDLData.wNumberOfModules;iNew++)
				{
					if(stDLOld.Module[iOld].wID==stDLData.Module[iNew].wID)
					{
						// VDIIɓW[
						if(stDLOld.Module[iOld].byVersion!=stDLData.Module[iNew].byVersion)
						{
							// o[WႤ̂ōXVĂ
							bUpdateFlg = true;
							// XVʒm
							TAG_TSCB_MODULE stNotify;
							stNotify.byComponentTag = byComponentTag;
							stNotify.wModuleID = stDLData.Module[iNew].wID;
							stNotify.byVersion = stDLData.Module[iNew].byVersion;
							stNotify.dwSize = stDLData.Module[iNew].dwSize;
							DoCallback(TSENG_CALLBACK_DATA_MODULE_UPDATE,wPID,&stNotify);
						} else {
							// o[WȂ̂ŁAXV͂ĂȂ(Xe[^Xƃobt@Rs[̂ŊJȂ)
							stDLData.Module[iNew] = stDLOld.Module[iOld];
							bFreeFlg = false;
						}
						bFindFlg = true;
					}
				}
				if(!bFindFlg)
				{
					// VDIIɃW[ԍ݂Ȃ̂ stDLOld.Module[iOld] ̓W[폜ꂽ
					DWORD dwLocationID=GetLocationID(stDLOld.byComponentTag,stDLOld.Module[iOld].wID);
					TsContentLocation::iterator itLocation=m_Content.find(dwLocationID);
					if(itLocation!=m_Content.end())
					{
#ifdef TSDATAMGR_DEBUG_PRINT
						{
							char dbg_buf[80];
							wsprintfA(dbg_buf,"DEL ModNo=0x%04X\n",stDLOld.Module[iOld].wID);
							OutputDebugStringA(dbg_buf);
						}
#endif
						TsContent::iterator itContent;
						for(itContent=itLocation->second.begin();itContent!=itLocation->second.end();itContent++)
						{
#ifdef TSDATAMGR_DEBUG_PRINT
							{
								char dbg_buf[80];
								wsprintfA(dbg_buf,"\tDelete Content \'%s\'\n",itContent->first.c_str());
								OutputDebugStringA(dbg_buf);
							}
#endif
							ContentBuffer_Free(itContent->second.pContent);
						}
						// W[菜
						m_Content.erase(itLocation);
					}
					// 폜ʒm
					TAG_TSCB_MODULE stNotify;
					stNotify.byComponentTag = byComponentTag;
					stNotify.wModuleID = stDLOld.Module[iOld].wID;
					stNotify.byVersion = stDLOld.Module[iOld].byVersion;
					stNotify.dwSize = stDLOld.Module[iOld].dwSize;
					DoCallback(TSENG_CALLBACK_DATA_MODULE_DELETE,wPID,&stNotify);
				}
				if(bFreeFlg)
				{
					// VDIIɕKvȂobt@NA
					DownloadBuffer_Free(stDLOld.Module[iOld].Status.pDownload);
				}
			}
			// XV
			it->second = stDLData;
		}
		
		if(bUpdateFlg)
		{
#ifdef TSDATAMGR_DEBUG_PRINT
			char buf[200];
			wsprintfA(buf,"DDI Update PID=0x%04X TransactionID=0x%08X DowoloadID=0x%08X BlockSize=%d\n",
				wPID,
				stDLData.dwTransactionID,stDLData.dwDownloadID,stDLData.wBlockSize
				);
			OutputDebugStringA(buf);
#endif
			for(int iModule=0;iModule<stDLData.wNumberOfModules;iModule++)
			{
#ifdef TSDATAMGR_DEBUG_PRINT
				wsprintfA(buf,"\tModNo=0x%04X Ver=%02X Size=%6d(%3d Blocks)",
					stDLData.Module[iModule].wID,stDLData.Module[iModule].byVersion,stDLData.Module[iModule].dwSize,stDLData.Module[iModule].wBlockNum
					);
				OutputDebugStringA(buf);
				if(stDLData.Module[iModule].Compress.bFlg)
				{
					wsprintfA(buf,"\tCompress Type=%d DecompressSize=%6d",
						stDLData.Module[iModule].Compress.byType,stDLData.Module[iModule].Compress.dwSize
						);
					OutputDebugStringA(buf);
				}
				OutputDebugStringA("\n");
#endif
			}
		}
	}
	return true;
}

// DSM-CC(DownloadDataBlock TableID=0x3C)̃pPbg
bool CTsDataMgr::OnDDB(WORD wPID,BYTE *pSection,int iPos,int iRemain,BYTE byComponentTag)
{
	CLock cDataLock(&m_DataLock);

	// MPEG-2 DSM-CC
	if(pSection[iPos+0]==0x11 && pSection[iPos+1]==0x03 && ( pSection[iPos+2]==0x10 && pSection[iPos+3]==0x03 ))
	{
		DWORD dwDownloadID = (DWORD)(pSection[iPos+4])<<24 | (DWORD)(pSection[iPos+5])<<16 | (DWORD)(pSection[iPos+6])<<8 | pSection[iPos+7];
		BYTE byAdaptationLength = pSection[iPos+9];
		WORD wMessageLength = (WORD)(pSection[iPos+10])<<8 | pSection[iPos+11];

		TsDownloadData::iterator it = m_DownloadData.find(wPID);
		if(it==m_DownloadData.end())
		{
			// PID̃_E[h񂪂܂s
			return true;
		}
		TAG_DownloadData &stDLData = it->second;		
		if(dwDownloadID!=stDLData.dwDownloadID)
		{
			// DownloadIDႤ
			return true;
		}

		iPos+=(byAdaptationLength+12);
		iRemain-=(byAdaptationLength+12);

		WORD wModuleID = (WORD)(pSection[iPos+0])<<8 | pSection[iPos+1];
		BYTE byModuleVersion = pSection[iPos+2];
		WORD wBlockNo = (WORD)(pSection[iPos+4])<<8 | pSection[iPos+5];		

		// ǂ̃W[
		int iModule=-1;
		for(int i=0;iModule==-1 && i<stDLData.wNumberOfModules;i++)
		{
			if(wModuleID==stDLData.Module[i].wID)
			{
				iModule=i;
			}
		}
		if(iModule==-1)
		{
			// ModuleID_E[hɂȂ
			return true;
		}
		if(stDLData.Module[iModule].Status.bCompleteFlg)
		{
			// Ƀ_E[hς݂̃f[^
			return TRUE;
		}
		if(wBlockNo>stDLData.Module[iModule].wBlockNum)
		{
			// ubNI[o[(G[f[^̃vȎzO)
			return true;
		}
		iPos+=6;
		iRemain-=6;

		// _E[hpȂȂm
		PBYTE &pDownload = stDLData.Module[iModule].Status.pDownload;
		if(!pDownload)
		{			
			if(!DownloadBuffer_Alloc(pDownload,stDLData.Module[iModule].wBlockNum*stDLData.wBlockSize))
			{
				// mۃG[
				return TRUE;
			}
		}

		// _E[hubNo^
		memcpy(pDownload+(wBlockNo*stDLData.wBlockSize),pSection+iPos,iRemain);
		stDLData.Module[iModule].Status.bBlock[wBlockNo]=true;		
		bool bCompleteFlg=true;
		for(int iBlock=0;bCompleteFlg && iBlock<stDLData.Module[iModule].wBlockNum;iBlock++)
		{
			if(!stDLData.Module[iModule].Status.bBlock[iBlock]) bCompleteFlg=false;
		}
		if(bCompleteFlg)
		{
			// _E[h
			stDLData.Module[iModule].Status.bCompleteFlg=true;			
			if(stDLData.Module[iModule].Compress.bFlg)
			{
				// kĂȂzlibWJ
				PBYTE pUncompress=NULL;
				if(DownloadBuffer_Alloc(pUncompress,stDLData.Module[iModule].Compress.dwSize))
				{
					z_stream zst;
					zst.zalloc = (alloc_func)0;
					zst.zfree = (free_func)0;
					zst.next_in = pDownload;
					zst.avail_in = stDLData.Module[iModule].dwSize;
					zst.next_out = pUncompress;
					zst.avail_out = stDLData.Module[iModule].Compress.dwSize;
					int z_err = inflateInit(&zst);
					if(z_err==Z_OK)
					{
						z_err = inflate(&zst, Z_FINISH);
						if(z_err!=Z_STREAM_END)
						{
							inflateEnd(&zst);
						}
						z_err = inflateEnd(&zst);
					}
					if(z_err==Z_OK)
					{
						ParseContent(stDLData.byComponentTag,stDLData.Module[iModule].wID,stDLData.Module[iModule].byVersion,pUncompress,stDLData.Module[iModule].Compress.dwSize);
					}
					DownloadBuffer_Free(pUncompress);
				}
			} else {
				// ̂܂
				ParseContent(stDLData.byComponentTag,stDLData.Module[iModule].wID,stDLData.Module[iModule].byVersion,pDownload,stDLData.Module[iModule].dwSize);
			}
			// J
			DownloadBuffer_Free(stDLData.Module[iModule].Status.pDownload);
			// W[_E[hʒm
			TAG_TSCB_MODULE stNotify;
			stNotify.byComponentTag = byComponentTag;
			stNotify.wModuleID = wModuleID;
			stNotify.byVersion = byModuleVersion;
			stNotify.dwSize = stDLData.Module[iModule].dwSize;
			DoCallback(TSENG_CALLBACK_DATA_MODULE_COMPLETE,wPID,&stNotify);
			// 
			int iDownloadComplete=0;
			for(int iModComp=0;iModComp<stDLData.wNumberOfModules;iModComp++)
			{
				if(stDLData.Module[iModComp].Status.bCompleteFlg) iDownloadComplete++;
			}
			if(iDownloadComplete==stDLData.wNumberOfModules)
			{
				// PID̑S_E[h
				stDLData.bAllDownloadComplete = true;
				// R|[lg_E[hʒm
				DoCallback(TSENG_CALLBACK_DATA_COMPONENT_COMPLETE,wPID,(LPVOID)byComponentTag);
			}
#ifdef TSDATAMGR_DEBUG_PRINT
			{
				char buf[200];
				wsprintfA(buf,"DDB Complete PID=0x%04X TransactionID=0x%08X DowoloadID=0x%08X Complete(%d/%d)\n",
					wPID,
					stDLData.dwTransactionID,stDLData.dwDownloadID,
					iDownloadComplete,stDLData.wNumberOfModules
					);
				OutputDebugStringA(buf);
				wsprintfA(buf,"\tModNo=0x%04X Ver=%02X Size=%6d(%3d Blocks)\n",
					stDLData.Module[iModule].wID,stDLData.Module[iModule].byVersion,stDLData.Module[iModule].dwSize,stDLData.Module[iModule].wBlockNum
					);
				OutputDebugStringA(buf);
				if(stDLData.bAllDownloadComplete)
				{
					wsprintfA(buf,"PID 0x%04X is Download Complete!\n",wPID);
					OutputDebugStringA(buf);
				}
			}
#endif			
		}
	}
	return true;
}

// obt@當Ps𔲂oA|C^i߂B
// ߂l͎擾oCgB}CiXȂf[^Ȃ
int CTsDataMgr::GetTextLine(PBYTE &pPtr,int &iRemain,PBYTE pBuf,int iBufSize)
{
	if(!pPtr || iRemain<=0) return -1;
	int i;
	for(i=0;i<iBufSize;i++)
	{
		pBuf[i]=*pPtr;
		pPtr++;
		iRemain--;
		// \0
		if(pBuf[i]=='\0')
		{
			iBufSize=i;
			return i;
		}
		// \r\n
		if(pBuf[i]=='\r' && *pPtr=='\n')
		{			
			iBufSize=i;
			pBuf[i]='\0';
			pPtr++;
			iRemain--;
			return i;
		}
		// \r or \n
		if(pBuf[i]=='\r' || pBuf[i]=='\n')
		{
			iBufSize=i;
			pBuf[i]='\0';
			return i;
		}
	}
	return i;
}

// wb_𕪉( <wb_> : <l> ̍\O )
// ߂l̓wb_̒(}CiX̓obt@s)
int CTsDataMgr::GetHeader(PBYTE pHeader,PBYTE pBufHeaderName,int iBufSize,PBYTE &pHeaderData)
{
	char *pSpliter = ": ";
	int iRet;
	int iHeaderLen = (int)strcspn((char*)pHeader,pSpliter);
	if(iHeaderLen+1 < iBufSize)
	{
		lstrcpynA((char*)pBufHeaderName,(char*)pHeader,iHeaderLen+1);
		iRet = iHeaderLen;
	} else {
		lstrcpynA((char*)pBufHeaderName,(char*)pHeader,iBufSize);
		iRet = -iHeaderLen;
	}
	int iSpliterLen=(int)strspn((char*)pHeader+iHeaderLen,pSpliter);
	pHeaderData=pHeader+iHeaderLen+iSpliterLen;
	return iRet;
}

// wb_̃f[^̑擾( <> = <l>)
int CTsDataMgr::GetHeaderAttribute(PBYTE pHeaderData,char* pAttrName,PBYTE pBufAttrData,int iBufSize)
{
	char *pSpliter = "\t =";
	char *pAttr=strstr((char*)pHeaderData,pAttrName);
	if(pAttr==NULL) return 0;
	int iAttrLen=(int)strcspn(pAttr,pSpliter);	
	char *pVal=pAttr+iAttrLen;
	pVal+=strspn(pVal,pSpliter);
	int iDataLen;
	if(pVal[0]=='\"')
	{
		pVal++;
		iDataLen = (int)strcspn(pVal,"\"");
	} else {
		iDataLen = (int)strcspn(pVal," \t");
	}
	if(iDataLen+1<iBufSize)
	{
		lstrcpynA((char*)pBufAttrData,pVal,iDataLen+1);
		return iDataLen;
	} else {
		lstrcpynA((char*)pBufAttrData,pVal,iBufSize);
		return -iDataLen;
	}
}

// _E[hRecwb_͂ĕۑ
// <ComponentTag>/<ModuleID>/<FileName(wb_ɋL)>
bool CTsDataMgr::ParseContent(BYTE byComponentTag,WORD wModuleID,BYTE byVersion,BYTE *pData,int iLen)
{
	PBYTE pTop = pData;

	BYTE buf[1024],pHeadName[1024],*pHeadData;
	BYTE pDelimiter[80]="",pCloseDelimiter[80]="";
	int size;

	size=GetTextLine(pTop,iLen,buf,1024);
	if(size>=0)
	{
		bool bMultipartFlg = false;
		if(GetHeader(buf,pHeadName,1024,pHeadData)>0)
		{
			if(lstrcmpiA((char*)pHeadName,"Content-Type")==0)
			{
				if(strstr((char*)pHeadData,"multipart/mixed"))
				{
					bMultipartFlg = true;
					// }`p[gW[ׁ̈Af~^擾
					if(GetHeaderAttribute(pHeadData,"boundary",pDelimiter+2,80-2)>0)
					{
						// --<f~^> p[gJn <f~^>-- t@CI
						pDelimiter[0]=pDelimiter[1]='-';
						lstrcpyA((char*)pCloseDelimiter,(char*)pDelimiter);
						lstrcatA((char*)pCloseDelimiter,"--");
					}
				}
			}
			if(bMultipartFlg)
			{
				// Jnʒu܂œǂݔ΂
				while(GetTextLine(pTop,iLen,buf,1024)>=0)
				{
					if(lstrcmpiA((char*)buf,(char*)pDelimiter)==0) break;
					if(lstrcmpiA((char*)buf,(char*)pCloseDelimiter)==0) break;
				}
				// ǂݍ݃[v
				do
				{
					// If~^ȂǂݍݏI
					if(lstrcmpiA((char*)buf,(char*)pCloseDelimiter)==0) break;
					// f~^ΊJn
					if(lstrcmpiA((char*)buf,(char*)pDelimiter)==0)
					{
						TAG_DownloadContent Content;
						ZeroMemory(&Content,sizeof(Content));
						Content.byVersion = byVersion;
						// wb_Jn
						while(GetTextLine(pTop,iLen,buf,1024)>=0)
						{
							if(buf[0]=='\0') break;
							if(GetHeader(buf,pHeadName,1024,pHeadData)>0)
							{
								if(lstrcmpiA((char*)pHeadName,"Content-Type")==0)
								{
									lstrcpynA(Content.szType,(char*)pHeadData,80);
								}
								if(lstrcmpiA((char*)pHeadName,"Content-Length")==0)
								{
									Content.iSize=atol((char*)pHeadData);
								}
								if(lstrcmpiA((char*)pHeadName,"Content-Encoding")==0)
								{
									lstrcpynA(Content.szEncoding,(char*)pHeadData,80);
								}
								if(lstrcmpiA((char*)pHeadName,"Content-Location")==0)
								{
									lstrcpynA(Content.szLocation,(char*)pHeadData,80);
								}
							}
						}
						// f[^
						StoreContent(byComponentTag,wModuleID,pTop,Content);
						// |C^i߂
						pTop+=Content.iSize;
						iLen-=Content.iSize;
					}
				} while(GetTextLine(pTop,iLen,buf,1024)>=0);
			}
		}
	}
	return true;
}

// Recێ
bool CTsDataMgr::StoreContent(BYTE byComponentTag,WORD wModuleID,BYTE *pData,TAG_DownloadContent &Content)
{
	CLock cDataLock(&m_DataLock);
	// ResourceList͕ۑȂ
	if(strstr(Content.szType,"application/X-arib-resourceList")) return false;
	// Location(Rec)Ȃ
	if(Content.szLocation[0]=='\0') return false;

	// Location(݂root̑Ό)
	DWORD dwLocationID=GetLocationID(byComponentTag,wModuleID);
	wsprintfA(Content.szPath,"%02X/%04X/%s",byComponentTag,wModuleID,Content.szLocation);
	TsContentLocation::iterator itLocation = m_Content.find(dwLocationID);
	if(itLocation==m_Content.end())
	{
		// VKP[V
		TsContent tmp;
		m_Content.insert(TsContentLocationPair(dwLocationID,tmp));
		itLocation = m_Content.find(dwLocationID);
		if(itLocation==m_Content.end())
		{
			// ǉs
			return false;
		}
	}
	TsContent &ContentMap = itLocation->second;
	TsContent::iterator itContent = ContentMap.find(Content.szLocation);
	if(itContent==ContentMap.end())
	{
		// VKRec
		ContentMap.insert(TsContentPair(Content.szLocation,Content));
		itContent = ContentMap.find(Content.szLocation);
		if(itContent==ContentMap.end())
		{
			// ǉs
			return false;
		}
	}
	TAG_DownloadContent &TargetContent = itContent->second;
	
	// ۑ
	if(!ContentBuffer_Alloc(TargetContent.pContent,TargetContent.iSize))
	{
		// mێs
		return false;
	}
	memcpy(TargetContent.pContent,pData,TargetContent.iSize);
	return true;
}

// wPIDɊ֘Af[^t@Cɏo͂
BOOL CTsDataMgr::OutputDataContent(char *lpszDir,WORD wPID)
{
	CLock cDataLock(&m_DataLock);
	TsDownloadData::iterator itDownload = m_DownloadData.find(wPID);
	if(itDownload == m_DownloadData.end()) return FALSE;

	TsContentLocation::iterator itLocation;
	for(itLocation=m_Content.begin();itLocation!=m_Content.end();itLocation++)
	{
		if(CompareComponentTag(itLocation->first,itDownload->second.byComponentTag))
		{
			TsContent::iterator itContent;
			for(itContent=itLocation->second.begin();itContent!=itLocation->second.end();itContent++)
			{
				
				char szPath[MAX_PATH],szTmp[MAX_PATH];
				lstrcpyA(szTmp,itContent->second.szPath);
				for(int i=0;i<lstrlenA(szTmp);i++)
				{
					if(szTmp[i]=='/') szTmp[i]='\\';
				}
				lstrcpyA(szPath,lpszDir);
				PathAppendA(szPath,szTmp);
				lstrcpyA(szTmp,szPath);
				PathRemoveFileSpecA(szTmp);				
				CFileRecord::CreateDirs(szTmp);
				// t@Cɏ
				FILE *fp=fopen(szPath,"wb");
				if(fp)
				{
					fwrite(itContent->second.pContent,itContent->second.iSize,1,fp);
					fclose(fp);
				}
			}
		}
	}
	return TRUE;
}

// f[^擾B pRetBufNull|C^nB
// FreeData() ŎgpɊJ
BOOL CTsDataMgr::GetData(char *szContent,PBYTE &pRetBuf, int &iRetSize)
{
	if(pRetBuf) return FALSE;
	CLock cDataLock(&m_DataLock);

	// fBNgƃt@Cɕ
	char szDir[80],szFile[80];
	BYTE byComponentTag;
	WORD wModuleID;
	char *p=strrchr(szContent,'/');
	if(p)
	{
		lstrcpyA(szDir,szContent);
		szDir[p-szContent]='\0';
		lstrcpyA(szFile,p+1);
	} else {
		szDir[0]='\0';
		lstrcpyA(szFile,szContent);
	}
	// Dir = <ComponentTag>/<ModuleID>
	p=strrchr(szDir,'/');
	if(p)
	{
		int tmp;
		sscanf(szDir,"%X",&tmp);
		byComponentTag = (BYTE)tmp;
		sscanf(p+1,"%X",&tmp);
		wModuleID = tmp;
	} else {
		byComponentTag = 0;
		wModuleID=0;
	}
	// ANZX
	return GetData(byComponentTag,wModuleID,szFile,pRetBuf,iRetSize);
}

// f[^擾B pRetBufNull|C^nB
// FreeData() ŎgpɊJ
BOOL CTsDataMgr::GetData(BYTE byComponentTag,WORD wModuleID,char *szFileName,PBYTE &pRetBuf, int &iRetSize)
{
	if(pRetBuf) return FALSE;
	CLock cDataLock(&m_DataLock);

	TsContentLocation::iterator itLocation = m_Content.find(GetLocationID(byComponentTag,wModuleID));
	if(itLocation==m_Content.end())
	{
		return FALSE;
	}
	TsContent::iterator itContent = itLocation->second.find(szFileName);
	if(itContent==itLocation->second.end())
	{
		return FALSE;
	}

	pRetBuf = new BYTE[itContent->second.iSize];
	if(!pRetBuf) return FALSE;
	memcpy(pRetBuf,itContent->second.pContent,itContent->second.iSize);
	iRetSize = itContent->second.iSize;
	return TRUE;
}

// GetDataŊmۂ̊J
void CTsDataMgr::FreeData(PBYTE &pRetBuf)
{
	if(pRetBuf)
	{
		delete []pRetBuf;
		pRetBuf=NULL;
	}
}